import unittest
from unittest import mock
from pynetgear import Netgear, Device


class TestGetAttachedDevices(unittest.TestCase):

    def test_noSignalType(self):
        spy = NetgearSpy(RESPONSE_NO_SIGNAL_TYPE)
        mocked_netgear = mock.Mock(wraps=spy)

        result = spy.get_attached_devices()
        assert result == [Device(signal=100, ip='192.168.1.4',
                                 name='MACBOOK-PRO', mac='80:E6:50:13:2B:E0',
                                 type=None, link_rate=0),
                          Device(signal=100, ip='192.168.1.18',
                                 name='RASPBERRYPI', mac='B8:27:EB:D9:05:E1',
                                 type=None, link_rate=0)]

    def test_withSignalType(self):
        spy = NetgearSpy(RESPONSE_WITH_SIGNAL_TYPE)
        mocked_netgear = mock.Mock(wraps=spy)

        result = spy.get_attached_devices()
        assert result == [Device(signal=88, ip='192.168.1.2',
                                 name='android-ada682e3ff4d6b20',
                                 mac='10:68:3F:AA:AA:AA',
                                 type='wireless', link_rate=72),
                          Device(signal=100, ip='192.168.1.3', name='odin',
                                 mac='C8:9C:DC:AA:AA:AA', type='wired',
                                 link_rate=None)]

    def test_invalidResponse(self):
        spy = NetgearSpy(RESPONSE_INVALID)
        mocked_netgear = mock.Mock(wraps=spy)
        result = spy.get_attached_devices()
        assert result is None

    def test_responseMissingSplitChar(self):
        spy = NetgearSpy(RESPONSE_MISSGING_SPLIT_CHAR)
        mocked_netgear = mock.Mock(wraps=spy)
        result = spy.get_attached_devices()
        assert result is None

    def test_responseUnknownDevice(self):
        spy = NetgearSpy(RESONSE_UNKOWN_DEVICE)
        mocked_netgear = mock.Mock(wraps=spy)
        result = spy.get_attached_devices()
        assert result == [Device(signal=100, ip='192.168.1.2',
                                 name='<unknown>',
                                 mac='10:68:3F:AA:AA:AA',
                                 type=None, link_rate=0)]


if __name__ == '__main__':
    unittest.main()


class NetgearSpy(Netgear):
    def __init__(self, response):
        super().__init__
        self.response = response

    def _make_request(self, action, message, try_login_after_failure=True):
        result = True
        response = self.response
        return result, response


RESPONSE_NO_SIGNAL_TYPE = "<NewAttachDevice>\
2@1;192.168.1.4;MACBOOK-PRO;80:E6:50:13:2B:E0\
@3;192.168.1.18;RASPBERRYPI;B8:27:EB:D9:05:E1\
</NewAttachDevice>"

RESPONSE_WITH_SIGNAL_TYPE = "<NewAttachDevice>\
2@1;192.168.1.2;android-ada682e3ff4d6b20;10:68:3F:AA:AA:AA;wireless;72;88\
@2;192.168.1.3;odin;C8:9C:DC:AA:AA:AA;wired;;100\
</NewAttachDevice>"

RESPONSE_INVALID = "INVALID"

RESPONSE_MISSGING_SPLIT_CHAR = "<NewAttachDevice>\
UNKNOWN#FORMATTING\
</NewAttachDevice>"

RESONSE_UNKOWN_DEVICE = "<NewAttachDevice>\
1@1;192.168.1.2;&lt;unknown&gt;;10:68:3F:AA:AA:AA\
</NewAttachDevice>"



from io import StringIO
import xml.etree.ElementTree as ET


LLL2 = u"""<?xml version="1.0" encoding="UTF-8"?>
<SOAP-ENV:Envelope
   xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"
   SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<SOAP-ENV:Body>
<m:GetAttachDeviceResponse xmlns:m="urn:NETGEAR-ROUTER:service:DeviceInfo:1">    <NewAttachDevice>6@1;192.168.0.1;GOOGLE-HOME-MINI;11:22:24;wireless;61;11;Allow@2;192.168.0.2;LALALA;11:22:33:44:55:66;wireless;121;11;Allow@3;192.168.0.9;;11:22:33:44:55:79;wireless;2;50;Allow@4;192.168.0.7;XXX;11:22:33:44:55:66;wireless;64;61;Allow@5;192.168.0.2;DDD;11:11:11:11:22:33;wireless;71;78;Allow@6;192.168.1.11;GFRTER;11:11:22:22:33:44;wireless;11;11;Allow</NewAttachDevice>
</m:GetAttachDeviceResponse>
<ResponseCode>000</ResponseCode>
</SOAP-ENV:Body>
</SOAP-ENV:Envelope>"""

LLL = u"""<?xml version="1.0" encoding="UTF-8"?>
<soap-env:Envelope
        xmlns:soap-env="http://schemas.xmlsoap.org/soap/envelope/"
        soap-env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"
        >
<soap-env:Body>
    <m:GetAttachDeviceResponse
        xmlns:m="urn:NETGEAT-ROUTER:service:DeviceInfo:1">
        <NewAttachDevice>15@1;192.168.10.2;DD-WRT;20:4E:7F:71:D5:AB;wired;;100;Allow@2;192.168.10.4;EmilieTmsiniPad;A8:FA:D8:3F:5F:71;wireless;52;60;Allow@3;192.168.10.5;WN2000RPTv2;4C:60:DE:E5:BB:DA;wireless;216;100;Allow@4;192.168.10.6;ESP_8102F6;BC:DD:C2:81:02:F6;wireless;72;75;Allow@5;192.168.10.8;Haralds-iPad;98:FE:94:7D:F2:DC;wireless;58;55;Allow@6;192.168.10.9;Living room;EC:66:D1:0C:C0:97;wireless;36;83;Allow@7;192.168.10.10;Apple-TV;1C:1A:C0:6C:49:E6;wireless;72;76;Allow@8;192.168.10.13;Fiskene;28:10:7B:05:16:DF;wireless;52;64;Allow@9;192.168.10.14;--;00:23:C1:14:10:01;wired;;100;Allow@10;192.168.10.16;sonoff_switch_1-2015;B4:E6:2D:24:E7:DF;wireless;12;55;Allow@11;192.168.10.19;pi-gw;B8:27:EB:E6:68:0C;wired;;100;Allow@12;192.168.10.20;DESKTOP-R5K1PV1;00:24:D7:64:13:28;wireless;117;60;Allow@13;192.168.10.23;7pgmini;00:24:D6:82:65:18;wireless;150;67;Allow@14;192.168.10.29;homeassistant;B8:27:EB:C0:6B:C8;wireless;72;84;Allow@15;192.168.10.201;NAS520;5C:F4:AB:65:7B:9A;wired;;100;Allow</NewAttachDevice>
    </m:GetAttachDeviceResponse>
    <ResponseCode>000</ResponseCode>
</soap-env:Body>
</soap-env:Envelope>"""


BBB = u"""@1;192.168.30.166;IPCAM;EC:3D:FD:F7:87:A2;wireless;50;56;Allow@2;192.168.30.21;ESP_EDE74A;5C:CF:7F:ED:E7:4A;wireless;70;82;Allow@3;192.168.30.186;IPHONE;3C:2E:FF:10:5C:BB;wireless;761;54;Allow@4;192.168.30.12;ESP_A76BF4;5C:CF:7F:A7:6B:F4;wireless;70;62;Allow@5;192.168.30.155;PHILIPS-HUE;00:17:88:2B:07:DF;wired;;100;Allow@6;192.168.30.182;SUPERIPAD;20:EE:28:3E:F3:B5;wireless;845;64;Allow@7;192.168.30.184;RUNES-MBP;5C:96:9D:70:1D:D7;wireless;439;62;Allow@8;192.168.30.11;ESP_055C8E;2C:3A:E8:05:5C:8E;wireless;50;80;Allow@9;192.168.30.3;ESP_05730E;2C:3A:E8:05:73:0E;wireless;50;70;Allow@10;192.168.30.4;ESP_EAB7F1;5C:CF:7F:EA:B7:F1;wireless;70;100;Allow@11;192.168.30.5;ESP_EAF11C;5C:CF:7F:EA:F1:1C;wireless;56;62;Allow@12;192.168.30.6;ESP_EAEAFC;5C:CF:7F:EA:EA:FC;wired;;100;Allow@13;192.168.30.17;ESP_A396B0;5C:CF:7F:A3:96:B0;wireless;70;88;Allow@14;192.168.30.18;ESP_A3A4E2;5C:CF:7F:A3:A4:E2;wireless;63;72;Allow@15;192.168.30.110;HP9856AF;EC:8E:B5:98:56:B0;wireless;56;62;Allow@16;192.168.30.187;APPLE-TV4K;DC:56:E7:40:2D:95;wireless;845;56;Allow@17;192.168.30.15;ESP_ACAEF5;5C:CF:7F:AC:AE:F5;wireless;56;84;Allow@18;192.168.30.189;LOCALHOST;FC:F1:36:EA:80:52;wired;;100;Allow@19;192.168.30.13;ESP_AF6760;5C:CF:7F:AF:67:60;wireless;56;70;Allow@20;192.168.30.2;ESP_AF6ECF;5C:CF:7F:AF:6E:CF;wireless;70;64;Allow@21;192.168.30.10;ESP_AF647A;5C:CF:7F:AF:64:7A;wireless;70;76;Allow@22;192.168.30.20;ESP_AF110C;5C:CF:7F:AF:11:0C;wireless;70;56;Allow@23;192.168.30.9;ESP_AF6831;5C:CF:7F:AF:68:31;wireless;56;50;Allow@24;192.168.30.28;GW-DCEFCABD27CF;DC:EF:CA:BD:27:CF;wired;;100;Allow@25;192.168.30.8;ESP_29078B;A0:20:A6:29:07:8B;wireless;70;100;Allow@26;192.168.30.180;IPI-6;9C:FC:01:1E:85:D9;wireless;56;64;Allow@27;192.168.30.105;HASSBIANS;4C:60:DE:71:8D:C2;wireless;140;100;Allow@28;192.168.30.7;ESP_FD81E5;5C:CF:7F:FD:81:E5;wireless;70;56;Allow@29;192.168.30.24;<unknown>;00:24:8D:F6:9E:EA;wireless;35;56;Allow@30;192.168.30.181;HONOR_9;5C:03:39:C7:53:06;wireless;285;60;Allow@31;192.168.30.19;ESP_15CAB6;A0:20:A6:15:CA:B6;wireless;70;60;Allow@32;192.168.30.16;ESP_1600D0;A0:20:A6:16:00:D0;wireless;70;88;Allow@33;192.168.30.148;NETTDISKEN;00:11:32:12:73:72;wired;;100;Allow@34;192.168.30.14;ESP_178871;A0:20:A6:17:88:71;wireless;38;40;Allow@35;192.168.30.44;IPCAM;48:02:2D:82:7E:29;wireless;70;70;Allow@36;192.168.30.145;HEIMEN;B8:27:EB:DA:CA:4A;wired;;100;Allow@37;192.168.30.23;ANDROID-E2A0AF84B028CB56;7C:B9:60:0C:68:2D;wireless;70;64;Allow@38;192.168.30.50;MOTIONEYEOS;B8:27:EB:DC:DE:AA;wireless;70;86;Allow@39;192.168.30.183;GAMPAD;64:20:0C:70:FD:1A;wireless;63;62;Allow"""


HHH = u""

UNKNOWN_DEVICE_DECODED = '<unknown>'
UNKNOWN_DEVICE_ENCODED = '&lt;unknown&gt;'


def main():
    devices = []
    # Netgear inserts a double-encoded value for "unknown" devices
    decoded = HHH.strip().replace(UNKNOWN_DEVICE_ENCODED,
                                        UNKNOWN_DEVICE_DECODED)

    if not decoded or decoded == "0":
        return devices

    entries = decoded.split("@")

    # First element is the total device count
    entry_count = None
    if len(entries) > 1:
        entry_count = _convert(entries.pop(0), int)

    if entry_count is not None and entry_count != len(entries):
        print(entry_count)
        print(len(entries))
        print(
            """Number of devices should \
                be: %d but is: %d""".format(entry_count, len(entries)))

    for entry in entries:
        info = entry.split(";")

        if len(info) == 0:
            continue

        # Not all routers will report those
        signal = None
        link_type = None
        link_rate = None
        allow_or_block = None

        if len(info) >= 8:
            allow_or_block = info[7]
        if len(info) >= 7:
            link_type = info[4]
            link_rate = _convert(info[5], int)
            signal = _convert(info[6], int)

        if len(info) < 4:
            print("Unexpected entry: %s".format(info))
            continue
        
        print(info)


def main_old():
    success, node = _find_node(
        LLL,
        u".//GetAttachDeviceResponse/NewAttachDevice")
    if not success:
        return None
    print(node)


def _find_node(response, xpath):
    it = ET.iterparse(StringIO(response))
    # strip all namespaces
    for _, el in it:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]
    node = it.root.find(xpath)
    if node is None:
        print("Error finding node in response: %s".format(response))
        return False, None

    return True, node


def _convert(value, to_type, default=None):
    """Convert value to to_type, returns default if fails."""
    try:
        return default if value is None else to_type(value)
    except ValueError:
        # If value could not be converted
        return default


if __name__ == '__main__':
    main()



from setuptools import setup

setup(name='pynetgear',
      version='0.4.1',
      description='Access Netgear routers using their SOAP API',
      url='http://github.com/MatMaul/pynetgear',
      download_url='http://github.com/MatMaul/pynetgear/archive/0.4.1.tar.gz',
      author='Paulus Schoutsen',
      author_email='Paulus@PaulusSchoutsen.nl',
      license='MIT',
      install_requires=['requests>=2.0'],
      packages=['pynetgear'],
      zip_safe=True)



"""Module to communicate with Netgear routers using the SOAP v2 API."""
from __future__ import print_function

from io import StringIO
from collections import namedtuple
import logging
import xml.etree.ElementTree as ET
from datetime import timedelta

import requests

DEFAULT_HOST = 'routerlogin.net'
DEFAULT_USER = 'admin'
DEFAULT_PORT = 5000
_LOGGER = logging.getLogger(__name__)

Device = namedtuple(
    "Device", ["signal", "ip", "name", "mac", "type", "link_rate",
               "allow_or_block", "device_type", "device_model",
               "ssid", "conn_ap_mac"])


class Netgear(object):
    """Represents a session to a Netgear Router."""

    def __init__(self, password=None, host=None, user=None, port=None,
                 ssl=False, url=None):
        """Initialize a Netgear session."""
        if not url and not host and not port:
            url = autodetect_url()

        if url:
            self.soap_url = url + "/soap/server_sa/"
        else:
            if not host:
                host = DEFAULT_HOST
            if not port:
                port = DEFAULT_PORT
            scheme = "https" if ssl else "http"
            self.soap_url = "{}://{}:{}/soap/server_sa/".format(scheme,
                                                                host, port)

        if not user:
            user = DEFAULT_USER

        self.username = user
        self.password = password
        self.port = port
        self.logged_in = False

    def login(self):
        """
        Login to the router.

        Will be called automatically by other actions.
        """
        _LOGGER.info("Login")

        body = LOGIN_BODY.format(username=self.username,
                                 password=self.password)

        success, _ = self._make_request("ParentalControl:1", "Authenticate",
                                        None, body, False)

        self.logged_in = success

        return self.logged_in

    def get_attached_devices(self):
        """
        Return list of connected devices to the router.

        Returns None if error occurred.
        """
        _LOGGER.info("Get attached devices")

        success, response = self._make_request(SERVICE_DEVICE_INFO,
                                               "GetAttachDevice")

        if not success:
            return None

        success, node = _find_node(
            response,
            ".//GetAttachDeviceResponse/NewAttachDevice")
        if not success:
            return None

        devices = []

        # Netgear inserts a double-encoded value for "unknown" devices
        decoded = node.text.strip().replace(UNKNOWN_DEVICE_ENCODED,
                                            UNKNOWN_DEVICE_DECODED)

        if not decoded or decoded == "0":
            return devices

        entries = decoded.split("@")

        # First element is the total device count
        entry_count = None
        if len(entries) > 1:
            entry_count = _convert(entries.pop(0), int)

        if entry_count is not None and entry_count != len(entries):
            _LOGGER.warning(
                """Number of devices should \
                 be: %d but is: %d""", entry_count, len(entries))

        for entry in entries:
            info = entry.split(";")

            if len(info) == 0:
                continue

            # Not all routers will report those
            signal = None
            link_type = None
            link_rate = None
            allow_or_block = None

            if len(info) >= 8:
                allow_or_block = info[7]
            if len(info) >= 7:
                link_type = info[4]
                link_rate = _convert(info[5], int)
                signal = _convert(info[6], int)

            if len(info) < 4:
                _LOGGER.warning("Unexpected entry: %s", info)
                continue

            ipv4, name, mac = info[1:4]

            devices.append(Device(signal, ipv4, name, mac,
                                  link_type, link_rate, allow_or_block,
                                  None, None, None, None))

        return devices

    def get_attached_devices_2(self):
        """
        Return list of connected devices to the router with details.

        This call is slower and probably heavier on the router load.

        Returns None if error occurred.
        """
        _LOGGER.info("Get attached devices 2")

        success, response = self._make_request(SERVICE_DEVICE_INFO,
                                               "GetAttachDevice2")
        if not success:
            return None

        success, devices_node = _find_node(
            response,
            ".//GetAttachDevice2Response/NewAttachDevice")
        if not success:
            return None

        xml_devices = devices_node.findall("Device")
        devices = []
        for d in xml_devices:
            ip = _xml_get(d, 'IP')
            name = _xml_get(d, 'Name')
            mac = _xml_get(d, 'MAC')
            signal = _convert(_xml_get(d, 'SignalStrength'), int)
            link_type = _xml_get(d, 'ConnectionType')
            link_rate = _xml_get(d, 'Linkspeed')
            allow_or_block = _xml_get(d, 'AllowOrBlock')
            device_type = _convert(_xml_get(d, 'DeviceType'), int)
            device_model = _xml_get(d, 'DeviceModel')
            ssid = _xml_get(d, 'SSID')
            conn_ap_mac = _xml_get(d, 'ConnAPMAC')
            devices.append(Device(signal, ip, name, mac, link_type, link_rate,
                                  allow_or_block, device_type, device_model,
                                  ssid, conn_ap_mac))

        return devices

    def get_traffic_meter(self):
        """
        Return dict of traffic meter stats.

        Returns None if error occurred.
        """
        _LOGGER.info("Get traffic meter")

        def parse_text(text):
            """
                there are three kinds of values in the returned data
                This function parses the different values and returns
                (total, avg), timedelta or a plain float
            """
            def tofloats(lst): return (float(t) for t in lst)
            try:
                if "/" in text:  # "6.19/0.88" total/avg
                    return tuple(tofloats(text.split('/')))
                elif ":" in text:  # 11:14 hr:mn
                    hour, mins = tofloats(text.split(':'))
                    return timedelta(hours=hour, minutes=mins)
                else:
                    return float(text)
            except ValueError:
                return None

        success, response = self._make_request(SERVICE_DEVICE_CONFIG,
                                               "GetTrafficMeterStatistics")
        if not success:
            return None

        success, node = _find_node(
            response,
            ".//GetTrafficMeterStatisticsResponse")
        if not success:
            return None

        return {t.tag: parse_text(t.text) for t in node}

    def _make_request(self, service, method, params=None, body="",
                      try_login_after_failure=True):
        """Make an API request to the router."""
        # If we are not logged in, the request will fail for sure.
        if not self.logged_in and try_login_after_failure:
            if not self.login():
                return False, ""

        headers = _get_soap_header(service, method)

        if not body:
            if not params:
                params = ""
            if isinstance(params, dict):
                map = params
                params = ""
                for k in map:
                    params += "<" + k + ">" + map[k] + "</" + k + ">\n"

            body = CALL_BODY.format(service=SERVICE_PREFIX + service,
                                    method=method, params=params)

        message = SOAP_REQUEST.format(session_id=SESSION_ID, body=body)

        try:
            req = requests.post(self.soap_url, headers=headers,
                                data=message, timeout=30, verify=False)

            success = _is_valid_response(req)

            if not success and try_login_after_failure:
                self.login()

                req = requests.post(self.soap_url, headers=headers,
                                    data=message, timeout=30, verify=False)

                success = _is_valid_response(req)

            return success, req.text

        except requests.exceptions.RequestException:
            _LOGGER.exception("Error talking to API")

            # Maybe one day we will distinguish between
            # different errors..
            return False, ""


def autodetect_url():
    """
    Try to autodetect the base URL of the router SOAP service.

    Returns None if it can't be found.
    """
    for url in ["http://routerlogin.net:5000", "https://routerlogin.net",
                "http://routerlogin.net"]:
        try:
            r = requests.get(url + "/soap/server_sa/",
                             headers=_get_soap_header("Test:1", "test"),
                             verify=False)
            if r.status_code == 200:
                return url
        except:
            pass

    return None


def _find_node(response, xpath):
    it = ET.iterparse(StringIO(response))
    # strip all namespaces
    for _, el in it:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]
    node = it.root.find(xpath)
    if node is None:
        _LOGGER.error("Error finding node in response: %s", response)
        return False, None

    return True, node


def _xml_get(e, name):
    """
    Returns the value of the subnode "name" of element e.

    Returns None if the subnode doesn't exist
    """
    r = e.find(name)
    if r is not None:
        return r.text
    return None


def _get_soap_header(service, method):
    action = SERVICE_PREFIX + service + "#" + method
    return {"SOAPAction": action}


def _is_valid_response(resp):
    return (resp.status_code == 200 and
            "<ResponseCode>000</ResponseCode>" in resp.text)


def _convert(value, to_type, default=None):
    """Convert value to to_type, returns default if fails."""
    try:
        return default if value is None else to_type(value)
    except ValueError:
        # If value could not be converted
        return default

SERVICE_PREFIX = "urn:NETGEAR-ROUTER:service:"
SERVICE_DEVICE_INFO = "DeviceInfo:1"
SERVICE_DEVICE_CONFIG = "DeviceConfig:1"

REGEX_ATTACHED_DEVICES = r"<NewAttachDevice>(.*)</NewAttachDevice>"

# Until we know how to generate it, give the one we captured
SESSION_ID = "A7D88AE69687E58D9A00"

SOAP_REQUEST = """<?xml version="1.0" encoding="utf-8" standalone="no"?>
<SOAP-ENV:Envelope xmlns:SOAPSDK1="http://www.w3.org/2001/XMLSchema"
  xmlns:SOAPSDK2="http://www.w3.org/2001/XMLSchema-instance"
  xmlns:SOAPSDK3="http://schemas.xmlsoap.org/soap/encoding/"
  xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
<SOAP-ENV:Header>
<SessionID>{session_id}</SessionID>
</SOAP-ENV:Header>
{body}
</SOAP-ENV:Envelope>
"""

LOGIN_BODY = """<SOAP-ENV:Body>
<Authenticate>
  <NewUsername>{username}</NewUsername>
  <NewPassword>{password}</NewPassword>
</Authenticate>
</SOAP-ENV:Body>"""

CALL_BODY = """<SOAP-ENV:Body>
<M1:{method} xmlns:M1="{service}">
{params}</M1:{method}>
</SOAP-ENV:Body>"""

UNKNOWN_DEVICE_DECODED = '<unknown>'
UNKNOWN_DEVICE_ENCODED = '&lt;unknown&gt;'
